// Portable Game Notation

// constraint: player_to_move = ply_number % 2 + 1

/*
note that the initial FEN could contain a initial move
number which should be used as an offset in the move list

[FEN "xx3oo/6o/7/7/7/7/o4xx x 0"]
                              ^ initial fullMove (or halfMove?), other FENs have both the fullmove AND halfmove
*/

import { Board } from '../board/board.js'
import { Move } from '../board/move.js'

export class PgnLoader /* extends Loader */ {

    parse(text) {
        const lines = text.split('\n')

        // tags
        const tags = []
        for (let line = lines.shift(); line.trim() != ''; line = lines.shift()) {
            const tag = line.match(/\[(.+)\s+\"(.+)\"\]/)

            const [, name, value] = tag
            tags[name] = value
        }

        let players = []
        if (tags.White && tags.Black) {
            players = [tags.White, tags.Black]
        } else {
            console.warn(`'PGN: player tags missing`)
        }

        if (!tags.FEN && tags.Setup) {
            console.warn(`'PGN: FEN tag missing while Setup = 1`)
        }
        if (tags.FEN && !tags.Setup) {
            console.warn(`'PGN: missing 'Setup' tag`)
        }

        // moves
        const line = lines.shift()
        const san = line.match(/\w+\.\s+(\w+\s+\w+)/g)

        let consistentOrdering = true
        const moves = san.map((move, step) => {
            const [fullMove, white, black] = move.split(' ')

            consistentOrdering = consistentOrdering && fullMove == step + 1

            return [white, black]
        }).flat()

        if (!consistentOrdering) {
            console.warn(`[PGN] inconsistent full move numbering`)
        }

        if (tags.FEN) {
            const [initialFen, color, fullMove, halfMove] = tags.FEN.split(' ')
            const game = this.buildBoards(initialFen, fullMove, moves, color)
            moves.unshift('START')

            return { moves, players, game }
        } else {
            // TODO check if first move makes sense against this board setup (e.g. black a6 = illegal because it cannot jump that far)
            // TODO detect that .fen and .pgns start with black (you can see this if the first halfMove for 'x' is odd)

            const [initialFen, color, fullMove, halfMove] = ['o5x/7/7/7/7/7/x5o o 1 1']
            const game = this.buildBoards(initialFen, fullMove, moves, color)
            moves.unshift('START')

            return { moves, players, game }
        }
    }

    buildBoards(initialFen, fullMove, san, color) {
        // initial board set up
        let board = { board: Board.fromFen(initialFen) }
        const boards = [board]

        for (const [halfMove, move] of Object.entries(san)) {

            // TODO test halfMove, fullMove and board against .fen for the same game
            const fullMove = Math.floor(halfMove / 2)
            board = this.getBoard(board, move, color)
            if (board == null) {
                console.warn(`invalid move '${move}' at move ${fullMove}`)
                break
            }
            boards.push(board)

            color = color == 'o' ? 'x' : 'o'
        }
        return boards
    }

    getBoard(board, move, color) {
        const { source, destination } = Move.fromSAN(move)

        if (!source && !destination) {
            return null
        }

        return {
            move,
            source,
            destination,
            board: this.renderMove(board, source, destination, color)
        }
    }

    renderMove(previousBoard, source, destination, color) {
        // deep copy
        const board = previousBoard.board.map(row => row.slice())

        // clone or jump
        if (!source && destination) {
            board[destination.row][destination.column] = color
        }

        if (source && destination) {
            board[source.row][source.column] = color
            board[destination.row][destination.column] = color

            // remove source if it is a jump
            const isJump =
                Math.abs(source.row - destination.row) > 1 ||
                Math.abs(source.column - destination.column) > 1

            if (isJump) {
                board[source.row][source.column] = ''
            }
        }

        // infect
        const neighbours = [
            { row: -1, column: -1 },
            { row: 0, column: -1 },
            { row: 1, column: -1 },
            { row: -1, column: 0 },
            { row: 1, column: 0 },
            { row: -1, column: 1 },
            { row: 0, column: 1 },
            { row: 1, column: 1 }
        ]

        const opposites = { x: 'o', o: 'x' }
        for (const neighbour of neighbours) {
            const row = neighbour.row + (destination ? destination.row : source.row)
            const column = neighbour.column + (destination ? destination.column : source.column)
            if (row >= 0 && row < 7 && column >= 0 && column < 7) {
                if (board[row][column] == opposites[color]) {
                    board[row][column] = color
                }
            }
        }

        return board
    }
}

